"""
   SDC Client model module
"""
from sqlalchemy import Column, String, DateTime, JSON
from sqlalchemy.orm import relationship
from sdc_helpers.models.model import Model
from sdc_helpers.models.subscription import Subscription

class Client(Model):
    """
       SDC Client model class
    """
    # pylint: disable=too-few-public-methods
    __tablename__ = 'clients'

    code = Column(String(255), unique=True)
    name = Column(String(255))
    properties = Column(JSON)
    deleted_at = Column(DateTime)
    updated_by = Column(String(255))
    subscriptions = relationship(Subscription)

    def __init__(self, **kwargs):
        """
            Creation of client model
        """
        super().__init__(**kwargs)
        self.code = kwargs.get('code')
        self.name = kwargs.get('name')
        self.last_login_at = self.convert_json_datetime(kwargs.get('last_login_at'))



"""
   SDC Process model module
"""
from sqlalchemy import Column, String, DateTime
from sdc_helpers.models.model import Model


class Process(Model):
    """
       SDC Process model class
    """
    # pylint: disable=too-few-public-methods
    __tablename__ = 'processes'

    slug = Column(String(255), unique=True)
    name = Column(String(255))
    deleted_at = Column(DateTime)

    def __init__(self, **kwargs):
        """
            Creation of process model
        """
        super().__init__(**kwargs)
        self.slug = kwargs.get('slug')
        self.name = kwargs.get('name')



"""
   SDC Base model module
"""
# coding=utf-8
# pylint: disable=invalid-name, unused-argument
import copy
import json
from datetime import datetime
from sqlalchemy import Column, DateTime, Integer, event
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import query_expression

Base = declarative_base()
models_namespace = 'sdc_helpers.models'
json_date_format = '%Y-%m-%d %H:%M:%S'

class Model(Base):
    """
       SDC base model class
    """
    # pylint: disable=too-few-public-methods, no-member
    __abstract__ = True
    original = None

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)

    def __init__(self, **kwargs):
        """
            Creation of abstract model
        """
        model_id = kwargs.get('id')
        if model_id is not None:
            self.id = model_id
        self.for_update = kwargs.get('for_update', False)
        self.created_at = self.convert_json_datetime(kwargs.get('created_at'))
        self.updated_at = self.convert_json_datetime(kwargs.get('updated_at'))
        properties = kwargs.get('properties')
        if properties is not None:
            if isinstance(properties, str):
                properties = json.loads(properties)
            self.properties = properties

    @declared_attr
    def for_update(self):
        """
            Declared abstract model attribute
        """
        return query_expression()

    def to_dict(self) -> dict:
        """
            Convert model object to dictionary
        """
        return {
            column.name: getattr(self, column.name, None)
            for column in self.__table__.columns
        }

    @staticmethod
    def convert_json_datetime(date_string):
        """
            Convert JSON string date to datetime object

            args:
                date_string(str): JSON date string
        """
        if date_string is None:
            return None

        return datetime.strptime(date_string, json_date_format)

@event.listens_for(Model, 'load', restore_load_context=True, propagate=True)
def receive_load(target, context):
    """
        Event listener triggered on first load of model

        args:
            target: Model instance
            context: Context data of the model
     """
    target.original = copy.deepcopy(target.to_dict())



"""
   SDC Audit model module
"""
from sqlalchemy import Column, Integer, String, JSON
from sdc_helpers.models.model import Model


class Audit(Model):
    """
       SDC Audit model class
    """
    # pylint: disable=too-few-public-methods, too-many-instance-attributes
    __tablename__ = 'audits'

    user_type = Column(String(255))
    user_id = Column(Integer)
    event = Column(String(255))
    auditable_type = Column(String(255))
    auditable_id = Column(Integer)
    old_values = Column(JSON)
    new_values = Column(JSON)
    url = Column(String(255))
    ip_address = Column(String(255))
    user_agent = Column(String(255))
    tags = Column(String(255))

    def __init__(self, **kwargs):
        """
            Creation of process model
        """
        super().__init__(**kwargs)
        self.slug = kwargs.get('user_type')
        self.name = kwargs.get('user_id')
        self.event = kwargs.get('event')
        self.user_type = kwargs.get('user_type')
        self.user_id = kwargs.get('user_id')
        self.auditable_type = kwargs.get('auditable_type')
        self.auditable_id = kwargs.get('auditable_id')
        self.old_values = kwargs.get('old_values')
        self.new_values = kwargs.get('new_values')



"""Extended DagFactory"""
# pylint: disable=unused-import
from typing import Any, Dict
import yaml
from airflow import DAG # don't remove this, required for DAG validation
from dagfactory import DagFactory

class ExtendedDagFactory(DagFactory):
    """
        extends the dagfactory with a custom yaml loader constructor

        Takes a YAML config and generates DAGs.

        :param config_filepath: the filepath of the DAG factory YAML config file.
            Must be absolute path to file.
    """
    # pylint: disable=super-init-not-called
    def __init__(self, config_filepath: str, loader: yaml.SafeLoader = None) -> None:
        """init"""
        if loader is not None:
            self.loader = loader
        else:
            self.loader = yaml.FullLoader

        self._validate_config_filepath(
            config_filepath=config_filepath
        )
        self.config_filepath: str = config_filepath

        self.config: Dict[str, Any] = self._load_config(
            config_filepath=config_filepath
        )

    def _load_config(self, config_filepath: str) -> Dict[str, Any]:
        """
        Loads YAML config file to dictionary
        :returns: dict from YAML config file
        """
        try:
            with open(config_filepath, "r") as file_:
                config: Dict[str, Any] = yaml.load(
                    file_, Loader=self.loader
                )
        except Exception as err:
            raise Exception("Invalid DAG Factory config file") from err

        return config



"""
   SDC Slack helper module
"""
import json
import os
import requests


class SlackHelper:
    """
       Slack helper class to send logs to Slack
    """
    critical_hook = None
    critical_color = '#ff0909'

    def __init__(self):
        self.critical_hook = os.getenv('SLACK_CRITICAL_HOOK')

    @staticmethod
    def send_message(*, hook: str, color: str, message: str):
        """
            Send a message to Slack channel

            args:
                hook (str): The Slack channel's hook URL endpoint
                color (str): Hex color code of the message
                value (str): The content of the message
        """
        data = {
            'attachments': [
                {
                    'title': os.getenv(
                        'AWS_PROCESS_NAME',
                        os.getenv(
                            'AWS_LAMBDA_FUNCTION_NAME',
                            'Unlabelled process'
                        )
                    ),
                    'text': message,
                    'color': color,
                    'mrkdwn_in': ['text']
                }
            ]
        }
        response = requests.post(
            url=hook,
            data=json.dumps(data)
        )

        response_text = response.text
        if response_text != 'ok':
            print('There was an issue posting to Slack: {error}'.format(error=response_text))

    def send_critical(self, *, message: str):
        """
            Send a critical message to the critical Slack channel

            args:
                value (str): The content of the message
        """
        self.send_message(
            hook=self.critical_hook,
            color=self.critical_color,
            message=message
        )



"""
   SDC Athena helper module
"""
import time
import boto3
from sdc_helpers import utils


def query(
        *,
        query_source: str,
        query_string: str,
        query_bucket: str = 'sdc-athena-queries',
        poll_count: int = 60,
        interval: float = 0.5
) -> list:
    """
        Run an Athena query and transform response into list of dictionaries
        Each dictionary is essentially a row of data

        args:
            query_source (str): Source folder in S3 bucket to store response data
            query_string (str): Athena Presto SQL query string
            query_bucket (str): S3 bucket results will be stored in
                                (Default - sdc-athena-queries)
            poll_count (int): The amount of times the function should check for query completion
                              (Default - 60)
            interval (float): The interval in seconds between polls (Default - 0.5)

        return:
            data (list) : A list of dictionary objects (rows of data)
    """
    client = boto3.client('athena')
    # pylint: disable=invalid-name,too-many-nested-blocks
    s3_location = "s3://{query_bucket}/{query_source}/".format(
        query_bucket=query_bucket,
        query_source=query_source
    )

    response = client.start_query_execution(
        QueryString=query_string,
        ResultConfiguration={
            'OutputLocation': s3_location,
        },
    )

    query_execution_id = response['QueryExecutionId']

    # we need to now poll the api for when the query has completed
    while poll_count > 0:
        response = client.get_query_execution(
            QueryExecutionId=query_execution_id
        )

        if response['QueryExecution']['Status']['State'] == 'FAILED':
            raise Exception(
                'Athena query failed with error: {error}'.format(
                    error=response['QueryExecution']['Status']
                )
            )

        if response['QueryExecution']['Status']['State'] \
                not in ['QUEUED', 'RUNNING']:
            next_token = None
            data = []
            columns = []

            while True:
                if next_token is not None:
                    response = client.get_query_results(
                        QueryExecutionId=query_execution_id,
                        NextToken=next_token
                    )
                else:
                    response = client.get_query_results(
                        QueryExecutionId=query_execution_id
                    )

                rows = utils.dict_query(dictionary=response, path='ResultSet.Rows')

                # Columns are only supplied on the first page
                if next_token is None:
                    columns = [column.get('VarCharValue') for column in rows[0].get('Data')]

                count = 0
                for row in rows:
                    if count > 0:
                        item = {}
                        value_index = 0
                        for value in row.get('Data'):
                            item[columns[value_index]] = value.get('VarCharValue')
                            value_index += 1

                        data.append(item)

                    count += 1

                next_token = response.get('NextToken')
                if next_token is None:
                    break

            return data

        poll_count -= 1
        time.sleep(interval)

    raise Exception('Athena query took too long')



"""Yaml Loader Constructors"""
import os
import yaml

def get_yaml_config(filepath: str, loader: yaml.SafeLoader = None) -> dict:
    """flexible yaml loader that takes defaults to safeloader"""
    if loader is None:
        loader = yaml.SafeLoader

    with open(filepath, 'r') as file_:
        return yaml.load(file_, Loader=loader)

class ExtendedSafeLoader(yaml.SafeLoader):
    """Extensible yaml.SafeLoader class where we add extra constructor"""

    # pylint: disable=too-many-ancestors
    def __init__(self, stream):
        """init ExtendedSafeLoader"""
        self._root = os.path.split(stream.name)[0]
        super().__init__(stream)

    def include(self, node):
        """macro to include !include function to yaml.load"""
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, 'r') as file_:
            return yaml.load(file_, Loader=ExtendedSafeLoader)

    def join(self, node):
        """macro to include !join function to yaml.load"""
        seq = self.construct_sequence(node)
        return ''.join([str(i) for i in seq])

    def load_sql(self, node):
        """macro to include !load_sql function to yaml.load"""
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, 'r') as file_:
            return file_.read()



"""
   SDC Utilities module
"""
import sys
import traceback

def extract_stack_trace() -> str:
    """Extracts full stacktrace from thown exception

    Returns:
        str: traceback of error thrown
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    stack_trace = traceback.format_exception(
        exc_type,
        exc_value,
        exc_traceback
    )
    return repr(stack_trace)

def parse_query_string_parameters(*, query_string_parameters: dict) -> dict:
    """
        Parse HTTP query string parameters into a more usable dictionary

        args:
            query_string_parameters (dict): The query string parameters from a request

        return:
            query_string_parameters (dict) : Returns the usable dictionary

    """
    parameters = {}
    for key, value in query_string_parameters.items():
        parts = key.replace(']', '').split('[')
        if len(parts) == 2:
            if not parameters.get(parts[0]):
                parameters[parts[0]] = {}
            parameters[parts[0]][parts[1]] = value
        else:
            parameters[parts[0]] = value

    return parameters


def dict_query(*, dictionary: dict, path: str, default=None):
    """
        Perform a 'dot notation' query on a dictionary

        args:
            dictionary (dict): The dictionary to query
            path (string): The dot notation query path
            default: The default value to return if key is not found

        return:
            value: Returns the value at the specified path

    """
    keys = path.split('.')
    value = None
    for key in keys:
        if value:
            if isinstance(value, list):
                if not key.isdigit():
                    return default

                value = list(value)[int(key)]
            else:
                value = value.get(key, None)
        else:
            value = dictionary.get(key, None)

        if not value:
            break

    return value if value is not None else default

def flatten_dict(dictionary: dict, separator: str = '.'):
    """
        Flatten a dictionary to a single level with keys separated by the provided separator

        args:
            dictionary (dict): The dictionary to manipulate
            separator: (str): The separator for the produced keys

            return:
                output_dictionary (dict): The manipulated dictionary

    """
    output_dictionary = {}

    def flatten(value, key_prefix: str = ''):
        """
            args:
                value: A dictionary node being processed
                key_prefix (str): The current key prefix through iterations
        """
        if isinstance(value, dict):
            for key in value:
                flatten(
                    value[key],
                    '{key_prefix}{key}{separator}'.format(
                        key_prefix=key_prefix,
                        key=key,
                        separator=separator
                    )
                )
        elif isinstance(value, list):
            for index, val in enumerate(value):
                flatten(
                    val,
                    '{key_prefix}{index}{separator}'.format(
                        key_prefix=key_prefix,
                        index=index,
                        separator=separator
                    )
                )
        else:
            output_dictionary[key_prefix[:-1]] = value

    flatten(dictionary)

    return output_dictionary



"""
    Set up the sdc_helpers package
"""
from setuptools import setup


def get_required(file_name):
    """
        Get Required Filename during setup
    """
    with open(file_name):
        return file_name.read().splitlines()


setup(
    name='sdc-helpers',
    packages=[
        'sdc_helpers',
        'sdc_helpers.models',
        'sdc_helpers.airflow_helpers',
        'sdc_helpers.yaml_helpers'
    ],
    install_requires=['sqlalchemy', 'pymysql', 'redis', 'boto3', 'requests', 'dag_factory'],
    description='Global SDC Helpers package',
    version='1.8.3',
    url='https://github.com/RingierIMU/sdc-global-all-helpers',
    author='Ringier South Africa',
    author_email='tools@ringier.co.za',
    keywords=['pip', 'helpers'],
    download_url='https://github.com/RingierIMU/sdc-global-all-helpers/archive/v1.8.1.zip'
)



"""
   SDC Service model module
"""
from sqlalchemy import Column, String, JSON, DateTime
from sqlalchemy.orm import relationship
from sdc_helpers.models.model import Model
from sdc_helpers.models.subscription import Subscription


class Service(Model):
    """
       SDC Service model class
    """
    # pylint: disable=too-few-public-methods
    __tablename__ = 'services'

    slug = Column(String(255), unique=True)
    name = Column(String(255))
    properties = Column(JSON)
    updated_by = Column(String(255))
    deleted_at = Column(DateTime)
    subscriptions = relationship(Subscription)

    def __init__(self, **kwargs):
        """
            Creation of service model
        """
        super().__init__(**kwargs)
        self.slug = kwargs.get('slug')
        self.name = kwargs.get('name')



"""
   SDC Redis helper module
"""
import os
import redis

class RedisHelper:
    """
       Redis helper class
    """
    redis_conn = None
    def __init__(self, host: str = None, port: str = None, db: int = None):
        # temporary fix - these should be passed at init time
        if host is None:
            host = os.getenv('REDIS_HOST', 'localhost')
        # temporary fix - these should be passed at init time
        if port is None:
            port = int(os.getenv('REDIS_PORT', '6379'))
        # temporary fix - these should be passed at init time
        if db is None:
            db = int(os.getenv('REDIS_DB', '0'))
        try:
            self.redis_conn = redis.Redis(
                host=host,
                port=port,
                db=db
            )
        except redis.ConnectionError as exception:
            print(exception)

    def __del__(self):
        self.redis_conn.connection_pool.disconnect()

    def redis_set(self, *, key: str, value: str, expiry: int = None):
        """
            Set a Redis key with the specified value

            args:
                key (str): The Redis key
                value (str): The value to set
                expiry (int): A TTL for the specified key
        """
        self.redis_conn.set(key, value, expiry)

    def redis_get(self, *, key: str) -> str:
        """
            Get a Redis key

            args:
                key (str): The Redis key

            return:
                value (str) : Returns the value for the specified key

        """
        return self.redis_conn.get(key)

    def redis_keys(self, *, pattern: str) -> list:
        """
            Get a list of Redis keys matching a pattern

            args:
                pattern (str): The pattern to search for

            return:
                keys (list) : Returns a list of keys found

        """
        keys = []
        for key in self.redis_conn.keys(pattern):
            keys.append(key.decode("utf-8"))

        return keys

    def redis_delete(self, *, keys) -> int:
        """
            Delete a Redis key

            args:
                key (str/list): One or many Redis keys

            return:
                result (int) : The number of keys deleted

        """
        if isinstance(keys, str):
            keys = [keys]
        return self.redis_conn.delete(*keys)

    def redis_flushall(self, asynchronous: bool = False):
        """Delete all keys in all databases on current host

        Args:
            asynchronous (bool): whether server can run flush asynchronously
        """
        self.redis_conn.flushall(asynchronous=asynchronous)

    def redis_flush_keys_matching(self, pattern: str = None):
        """
            Delete all keys matching pattern in database.
            Lookups up keys using SCAN and then mass deletes using
            redis_delete.

        Args:
            pattern (str, optional): A redis format pattern for matching keys. Defaults to None.

        Raises:
            Exception: [description]
        """
        if pattern is None:
            raise Exception((
                "No pattern given. To delete all keys, "
                "pattern = '*' or use a more specific pattern to match keys with "
            ))

        keys = list(self.redis_conn.scan_iter(pattern))
        self.redis_conn.delete(*keys)

    def redis_hexists(self, *, hashkey: str):
        """Check hashkey exists in store"""
        return self.redis_conn.exists(hashkey)

    def redis_h_check_key_exists(self, *, hashkey: str, key: str):
        """Check key exists in haskey in store"""
        return self.redis_conn.hexists(hashkey, key)

    def redis_hset(
            self,
            *,
            hashkey: str,
            key: str = None,
            value: str = None,
            dict_obj: dict = None
    ):
        """
            Create or Overwrite dictionary hash store.
            Either takes a dict_obj or (key, value) pair.
        """
        if dict_obj is not None:
            self.redis_conn.hset(name=hashkey, mapping=dict_obj)
        elif key is not None and value is not None:
            self.redis_conn.hset(name=hashkey, key=key, value=value)
        else:
            raise ValueError("Expected either dict_obj or key & value")

    def redis_hmget(self, *, hashkey: str, keys: list):
        """Read values for set of keys from hash store"""
        return self.redis_conn.hmget(name=hashkey, keys=keys)

    def redis_hget(self, *, hashkey: str, key: str):
        """Read values for a given key from hash store"""
        return self.redis_conn.hget(name=hashkey, key=key)

    def redis_hmgetall(self, *, hashkey: str):
        """Read complete dict object from hash store"""
        return self.redis_conn.hgetall(name=hashkey)



"""
   SDC subscription model module
"""
from sqlalchemy import Column, Integer, ForeignKey, JSON, String, DateTime
from sqlalchemy.orm import relationship, query_expression
from sdc_helpers.models.model import Model, models_namespace


class Subscription(Model):
    """
       SDC Subscription model class
    """
    # pylint: disable=too-few-public-methods
    __tablename__ = 'subscriptions'

    client_id = Column(Integer, ForeignKey('clients.id'), primary_key=True)
    service_id = Column(Integer, ForeignKey('services.id'), primary_key=True)
    properties = Column(JSON)
    updated_by = Column(String(255))
    deleted_at = Column(DateTime)
    properties_node = query_expression()

    client = relationship(
        '{models_namespace}.client.Client'.format(
            models_namespace=models_namespace
        )
    )
    service = relationship(
        '{models_namespace}.service.Service'.format(
            models_namespace=models_namespace
        )
    )

    def __init__(self, **kwargs):
        """
            Creation of subscription model
        """
        super().__init__(**kwargs)
        self.client_id = kwargs.get('client_id')
        self.service_id = kwargs.get('service_id')



"""
   SDC Query helper module
"""
import copy
import inspect
import os
import json
from datetime import datetime
import boto3
from sqlalchemy import create_engine, and_, literal
from sqlalchemy.orm import sessionmaker, undefer, Query
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.orm import with_expression
import sdc_helpers.utils as utils
import sdc_helpers.decorators as decorators
from sdc_helpers.models.audit import Audit
from sdc_helpers.models.client import Client
from sdc_helpers.models.model import Model
from sdc_helpers.models.process import Process
from sdc_helpers.models.service import Service
from sdc_helpers.models.subscription import Subscription
from sdc_helpers.redis_helper import RedisHelper


class QueryHelper:
    """
        Query helper class
    """
    # pylint: disable=too-few-public-methods, singleton-comparison, no-member

    session = None
    redis_helper = None
    mysql_date_format = '%Y-%m-%d %H:%M:%S'
    query_helper_caller = None
    process_id = None

    def __init__(self, **kwargs):
        """
            Init a new QueryHelper obj
            kwargs:
                rds_config (dict) : config arguments for mysql connector
                redis_config (dict) : config arguments for redis connector
        """
        caller_locals = inspect.currentframe().f_back.f_locals
        if caller_locals.get('self') is not None:
            caller_module = caller_locals['self'].__module__
            caller_class = caller_locals['self'].__class__.__name__
        else:
            caller_module = caller_locals.get('__module__')
            caller_class = caller_locals.get('__qualname__')

        self.query_helper_caller = caller_module

        if caller_class is not None:
            self.query_helper_caller = '{caller_module}.{caller_class}'.format(
                caller_module=caller_module,
                caller_class=caller_class
            )

        # get rds config
        rds_config = kwargs.get('rds_config', {})
        rds_host = rds_config.get('host', None)
        rds_user = rds_config.get('username', None)
        rds_password = rds_config.get('password', None)
        rds_port = rds_config.get('port', None)
        rds_db = rds_config.get('db', None)

        # get redis config
        redis_config = kwargs.get('redis_config', {})
        redis_host = redis_config.get('host', None)
        redis_port = redis_config.get('port', None)
        redis_db = redis_config.get('db', None)
        # temporary fix - these should be passed at init time
        if rds_host is None:
            rds_host = os.getenv('RDS_HOST', 'localhost')
        # temporary fix - these should be passed at init time
        if rds_user is None:
            rds_user = os.getenv('RDS_USERNAME', 'root')
        # temporary fix - these should be passed at init time
        if rds_password is None:
            rds_password = os.getenv('RDS_PASSWORD')
        # temporary fix - these should be passed at init time
        if rds_port is None:
            rds_port = int(os.getenv('RDS_PORT', '3306'))
        # temporary fix - these should be passed at init time
        if rds_db is None:
            rds_db = os.getenv('RDS_DB_NAME', 'sdc')

        user_password = rds_user
        if rds_password is not None:
            user_password = '{user}:{password}'.format(
                user=rds_user,
                password=rds_password
            )

        engine = create_engine(
            'mysql+pymysql://{user_password}@{host}:{port}/{db}'.format(
                user_password=user_password,
                host=rds_host,
                port=rds_port,
                db=rds_db
            )
        )

        self.session = sessionmaker(bind=engine)()

        # instantiate a redis helper
        self.redis_helper = RedisHelper(host=redis_host, port=redis_port, db=redis_db)

    def __del__(self):
        self.session.close()
        del self.redis_helper

    def get_query(
            self,
            model_class,
            for_update: bool = False,
            no_wait: bool = False
    ) -> Query:
        """
            Get a query for a model class

            args:
                model_class: The model class to query on
                for_update (bool): Reserve the model for exclusive use until
                                   commit or session is closed - Default False
                no_wait (bool): Don't wait around for a model if it is locked
            return:
                query (Query) : Return a SQLAlchemy Query object

        """
        # Check that the supplied class has an ancestor which is our base Model
        if Model not in model_class.__mro__:
            raise Exception('Please supply a SQLAlchemy model class to build a query')

        query = self.session.query(model_class).options(
            with_expression(
                model_class.for_update,
                literal(for_update)
            )
        )

        if for_update:
            query = query.with_for_update(nowait=no_wait)

        return query

    @decorators.query_exception_handler()
    def get_clients(self) -> list:
        """
            Get all the clients from the database

        return:
            clients (list) : Returns a list of all active Client models from the database

        """
        clients = self.get_query(Client).filter(
            Client.deleted_at == None
        ).all()

        return clients

    @decorators.query_exception_handler()
    def get_client(
            self,
            *,
            from_cache: bool = True,
            **kwargs
    ) -> Client:
        """
            Get the specified client from cache or database

            args:
                from_cache (bool): Retrieve the client from cache - Default True
                kwargs (dict):
                    api_key_id (str): The AWS API Gateway API Key Id
                    client_id (int): ID of the client in the database

            return:
                client (Client) : Returns the specified Client model

        """
        api_key_id = kwargs.get('api_key_id')
        client_id = kwargs.get('client_id')

        if not api_key_id and not client_id:
            raise Exception('ClientError: api_key_id or id is required for this function')

        if api_key_id and client_id:
            raise Exception(
                'ClientError: Only one of api_key_id or id should be specified for this function'
            )

        if api_key_id:
            client_redis_key = 'client-{api_key_id}'.format(api_key_id=api_key_id)
        else:
            client_redis_key = 'client-{client_id}'.format(client_id=client_id)

        client_redis = self.redis_helper.redis_get(key=client_redis_key)

        if (
                not from_cache or
                not client_redis
        ):
            query = self.get_query(Client)

            if api_key_id:
                client = boto3.client('apigateway')
                api_key = client.get_api_key(apiKey=api_key_id)

                if 'tags' not in api_key or 'client_code' not in api_key['tags']:
                    raise Exception(
                        ('ClientError: client_code not set up for this API key. '
                         'Please contact support'
                        )
                    )

                client = query.filter(
                    and_(
                        Client.code == utils.dict_query(
                            dictionary=api_key,
                            path='tags.client_code'
                        ),
                        Client.deleted_at == None
                    )
                ).first()
            else:
                client = query.filter(
                    and_(
                        Client.id == client_id,
                        Client.deleted_at == None
                    )
                ).first()

            if client is not None:
                self.redis_helper.redis_set(
                    key=client_redis_key,
                    value=json.dumps(client.to_dict(), default=str)
                )
        else:
            client_dict = json.loads(client_redis)
            client = Client(**client_dict)

        return client

    @decorators.query_exception_handler()
    def get_services(self) -> list:
        """
            Get all the services from the database

        return:
            services (list) : Returns a list of all active Service models from the database

        """
        services = self.get_query(Service).filter(
            Service.deleted_at == None
        ).all()

        return services

    @decorators.query_exception_handler()
    def get_service(self, *, slug: str, from_cache: bool = True) -> Service:
        """
            Get the specified service from cache or database

            args:
                slug (str): slug of the service in the database
                from_cache (bool): Retrieve the service from cache - Default True

            return:
                service (Service) : Returns the specified Service model

        """
        service_redis_key = 'service-{slug}'.format(slug=slug)
        service_redis = self.redis_helper.redis_get(key=service_redis_key)

        if (
                not from_cache or
                not service_redis
        ):
            service = self.get_query(Service).filter(
                and_(
                    Service.slug == slug,
                    Service.deleted_at == None
                )
            ).first()

            if service:
                self.redis_helper.redis_set(
                    key=service_redis_key,
                    value=json.dumps(service.to_dict(), default=str)
                )
        else:
            service_dict = json.loads(service_redis)
            service = Service(**service_dict)

        return service

    @decorators.query_exception_handler()
    def get_subscription(
            self,
            *,
            client_id: int,
            service_id: int,
            from_cache: bool = True,
            for_update: bool = False
    ) -> Subscription:
        """
            Get the specified subscription from cache or database

            args:
                client_id (id): client_id of the subscription in the database
                service_id (id): service_id of the subscription in the database
                from_cache (bool): Retrieve the subscription from cache - Default True
                for_update (bool): Reserve the subscription for exclusive use until
                                   commit or session is closed - Default False

            return:
                subscription (Subscription) : Returns the specified Subscription model

        """
        subscription_redis_key = (
            'subscription-{client_id}-{service_id}'.format(
                client_id=client_id,
                service_id=service_id
            )
        )
        subscription_redis = self.redis_helper.redis_get(
            key=subscription_redis_key
        )

        if (
                not from_cache or
                not subscription_redis
        ):
            query = self.get_query(Subscription, for_update).filter(
                and_(
                    Subscription.client_id == client_id,
                    Subscription.service_id == service_id,
                    Subscription.deleted_at == None
                )
            )

            subscription = query.first()

            if subscription:
                self.redis_helper.redis_set(
                    key=subscription_redis_key,
                    value=json.dumps(subscription.to_dict(), default=str)
                )
        else:
            subscription_dict = json.loads(subscription_redis)
            subscription = Subscription(**subscription_dict)

        return subscription

    @decorators.query_exception_handler()
    def get_subscriptions(self, *, service_id: int) -> list:
        """
            Get all the specified service's subscriptions from the database

            args:
                service_id (id): service_id of the subscription in the database

            return:
                subscriptions (list) : Returns the specified service's Subscription models

        """
        return self.get_query(Subscription, for_update=True).filter(
            and_(
                Subscription.service_id == service_id,
                Subscription.deleted_at == None
            )
        ).options(
            undefer('properties')
        ).all()

    @decorators.query_exception_handler()
    def get_process(self, *, slug: str, from_cache: bool = True) -> Process:
        """
            Get the specified process from cache or database

            args:
                slug (str): slug of the process in the database
                from_cache (bool): Retrieve the process from cache - Default True

            return:
                process (Process) : Returns the specified Process model

        """
        process_redis_key = 'process-{slug}'.format(slug=slug)
        process_redis = self.redis_helper.redis_get(key=process_redis_key)

        if (
                not from_cache or
                not process_redis
        ):
            process = self.get_query(Process).filter(
                and_(
                    Process.slug == slug,
                    Process.deleted_at == None
                )
            ).first()

            if process:
                self.redis_helper.redis_set(
                    key=process_redis_key,
                    value=json.dumps(process.to_dict(), default=str)
                )
        else:
            process_dict = json.loads(process_redis)
            process = Process(**process_dict)

        return process

    @decorators.query_exception_handler()
    def update(self, *, model: Model, audit: bool = True):
        """
            Commit the model's changes to the database

            Note:
                If any other changes have been made to other models, those would
                be committed too. It is the developer's responsibility to ensure
                that their workflow doesn't run into this issue

            args:
                model (Model): The model being updated
                audit (bool): Whether to create an audit for the update or not
        """
        if not model.for_update:
            raise Exception('ServerError: Cannot update a read only model')

        if hasattr(model, 'properties'):
            # If properties have only been partially modified,
            # e.g `subscription.properties['node1'][0][node2] = value`,
            # the properties attribute is not marked as dirty and we
            # need to mark it as such to cause an update in the database
            flag_modified(model, 'properties')

        now = datetime.strftime(
            datetime.now(),
            self.mysql_date_format
        )

        old_values = {}
        new_values = {}

        ignore_attributes = ['updated_at']

        original = set(utils.flatten_dict(model.original).items())
        new = set(utils.flatten_dict(model.to_dict()).items())

        for key, value in original.difference(new):
            if key in ignore_attributes:
                continue
            old_values[key] = value

        for key, value in new.difference(original):
            if key in ignore_attributes:
                continue
            new_values[key] = value

        if bool(old_values) or bool(new_values):
            if audit:
                if self.process_id is None:
                    raise Exception('Server Error: Process not set in QueryHelper for audits')

                fields = {
                    'user_type': 'Sdc\\Models\\Process',
                    'user_id': self.process_id,
                    'event': 'updated',
                    'auditable_type': 'Sdc\\Models\\{class_name}'.format(
                        class_name=model.__class__.__name__
                    ),
                    'auditable_id': model.id,
                    'old_values': old_values,
                    'new_values': new_values,
                    'created_at': now,
                    'updated_at': now
                }

                audit = Audit(**fields)
                self.session.add(audit)

            model.updated_at = now

            if hasattr(model, 'updated_by') and self.query_helper_caller is not None:
                model.updated_by = self.query_helper_caller

            self.session.commit()
            model.original = copy.deepcopy(model.to_dict())



"""
   SDC decorators module
"""
# pylint: disable=raise-missing-from
from functools import wraps
from sqlalchemy.exc import SQLAlchemyError
from sdc_helpers.slack_helper import SlackHelper
from sdc_helpers.utils import extract_stack_trace


def query_exception_handler(exceptions: tuple = (SQLAlchemyError, )):
    """
        Decorator - handling SqlAlchemy specific exceptions

        args:
            exceptions (Exception): List of exceptions to catch

        return:
            Wrapped function's response
    """
    def query_exception_decorator(function):
        @wraps(function)
        def func_with_exceptions(*args, **kwargs):
            """
                Wrapper function to decorate function with
            """
            try:
                return function(*args, **kwargs)
            except exceptions as ex:
                exception_class = ex.__class__
                if SQLAlchemyError in exception_class.__mro__:
                    message = 'Server Error: ({ex_class_module}.{ex_class_name}) {ex}'.format(
                        ex_class_module=ex.__class__.__module__,
                        ex_class_name=ex.__class__.__name__,
                        ex=str(ex.orig)
                    )
                else:
                    message = 'Server Error: {ex}'.format(
                        ex=str(ex)
                    )

                raise Exception(message)

        return func_with_exceptions

    return query_exception_decorator

def general_exception_handler(
        logger: SlackHelper,
        exceptions: tuple,
        should_raise: bool = True
    ):
    """
        Decorator - general handling for exceptiions which logs the exception using logger
        and optiionally can raise the error as well.

        Args:
            logger (SlackHelper): logger object of type SlackHelper
            exceptions (tuple): List of exceptions to catch
            should_raise (bool, optional): Should error be raised. Defaults to True.

        return:
            Wrapped function's response
    """
    def general_exception_decorator(function):

        @wraps(function)
        def func_with_exceptions(*args, **kwargs):
            """
                Wrapper function to decorate function with
            """
            try:
                return function(*args, **kwargs)

            except exceptions as ex:

                # get stack trace
                stack_trace = extract_stack_trace()
                # reformat log
                reformatted_log_msg = (
                    'Server Error: {ex}'.format(ex=stack_trace)
                )

                # emit log
                logger.send_critical(message=reformatted_log_msg)

                # raise error
                if should_raise:
                    raise ex

        return func_with_exceptions

    return general_exception_decorator
